"""Summarize the values in the columns of a columnar file."""

import pandas as pd
from hed.tools.analysis.tabular_summary import TabularSummary
from remodeler.operations.base_op import BaseOp
from remodeler.operations.base_summary import BaseSummary


class SummarizeColumnValuesOp(BaseOp):
    """Summarize the values in the columns of a columnar file.

    Required remodeling parameters:
        - **summary_name** (*str*): The name of the summary.
        - **summary_filename** (*str*): Base filename of the summary.

    Optional remodeling parameters:
        - **append_timecode** (*bool*): (**Optional**: Default False) If True append timecodes to the summary filename.
        - **max_categorical** (*int*): Maximum number of unique values to include in summary for a categorical column.
        - **skip_columns** (*list*):  Names of columns to skip in the summary.
        - **value_columns** (*list*): Names of columns to treat as value columns rather than categorical columns.
        - **values_per_line** (*int*): The number of values output per line in the summary.

    The purpose is to produce a summary of the values in a tabular file.

    """

    NAME = "summarize_column_values"

    PARAMS = {
        "type": "object",
        "properties": {
            "summary_name": {"type": "string", "description": "Name to use for the summary in titles."},
            "summary_filename": {"type": "string", "description": "Name to use for the summary file name base."},
            "append_timecode": {
                "type": "boolean",
                "description": "If true, the timecode is appended to the base filename so each run has a unique name.",
            },
            "max_categorical": {
                "type": "integer",
                "description": "Maximum number of unique column values to show in text description.",
            },
            "skip_columns": {
                "type": "array",
                "description": "List of columns to skip when creating the summary.",
                "items": {"type": "string"},
                "minItems": 1,
                "uniqueItems": True,
            },
            "value_columns": {
                "type": "array",
                "description": "Columns to be annotated with a single HED annotation and placeholder.",
                "items": {"type": "string"},
                "minItems": 1,
                "uniqueItems": True,
            },
            "values_per_line": {"type": "integer", "description": "Number of items per line to display in the text file."},
        },
        "required": ["summary_name", "summary_filename"],
        "additionalProperties": False,
    }

    SUMMARY_TYPE = "column_values"
    VALUES_PER_LINE = 5
    MAX_CATEGORICAL = 50

    def __init__(self, parameters):
        """Constructor for the summarize column values operation.

        Parameters:
            parameters (dict): Dictionary with the parameter values for required and optional parameters.

        """

        super().__init__(parameters)
        self.summary_name = parameters["summary_name"]
        self.summary_filename = parameters["summary_filename"]
        self.append_timecode = parameters.get("append_timecode", False)
        self.max_categorical = parameters.get("max_categorical", float("inf"))
        self.skip_columns = parameters.get("skip_columns", [])
        self.value_columns = parameters.get("value_columns", [])
        self.values_per_line = parameters.get("values_per_line", self.VALUES_PER_LINE)

    def do_op(self, dispatcher, df, name, sidecar=None) -> pd.DataFrame:
        """Create a summary of the column values in df.

        Parameters:
            dispatcher (Dispatcher): Manages the operation I/O.
            df (DataFrame): The DataFrame to be remodeled.
            name (str):  Unique identifier for the dataframe -- often the original file path.
            sidecar (Sidecar or file-like): Not needed for this operation.

        Returns:
            DataFrame: A copy of df.

        Side effect:
            Updates the relevant summary.

        """

        df_new = df.copy()
        summary = dispatcher.summary_dicts.get(self.summary_name, None)
        if not summary:
            summary = ColumnValueSummary(self)
            dispatcher.summary_dicts[self.summary_name] = summary
        summary.update_summary({"df": dispatcher.post_proc_data(df_new), "name": name})
        return df_new

    @staticmethod
    def validate_input_data(parameters) -> list:
        """Additional validation required of operation parameters not performed by JSON schema validator."""
        return []


class ColumnValueSummary(BaseSummary):
    """Manager for summaries of column contents for columnar files."""

    def __init__(self, sum_op):
        """Constructor for column value summary manager.

        Parameters:
            sum_op (SummarizeColumnValuesOp): Operation associated with this summary.

        """
        super().__init__(sum_op)

    def update_summary(self, new_info):
        """Update the summary for a given tabular input file.

        Parameters:
            new_info (dict):  A dictionary with the parameters needed to update a summary.

        Notes:
            - The summary information is kept in separate TabularSummary objects for each file.
            - The summary needs a "name" str and a "df" .

        """
        name = new_info["name"]
        if name not in self.summary_dict:
            self.summary_dict[name] = TabularSummary(
                value_cols=self.op.value_columns, skip_cols=self.op.skip_columns, name=name
            )
        self.summary_dict[name].update(new_info["df"])

    def get_details_dict(self, tabular_summary) -> dict:
        """Return a dictionary with the summary contained in a TabularSummary.

        Parameters:
            tabular_summary (TabularSummary): The tabular summary object.

        Returns:
            dict: Dictionary with the information suitable for extracting printout.

        """
        this_summary = tabular_summary.get_summary(as_json=False)
        unique_counts = [(key, len(count_dict)) for key, count_dict in this_summary["Categorical columns"].items()]
        this_summary["Categorical counts"] = dict(unique_counts)
        for key, dict_entry in this_summary["Categorical columns"].items():
            num_disp, sorted_tuples = ColumnValueSummary.sort_dict(dict_entry, reverse=True)
            this_summary["Categorical columns"][key] = dict(sorted_tuples[: min(num_disp, self.op.max_categorical)])
        return {
            "Name": this_summary["Name"],
            "Total events": this_summary["Total events"],
            "Total files": this_summary["Total files"],
            "Files": list(this_summary["Files"].keys()),
            "Specifics": {
                "Value columns": list(this_summary["Value columns"]),
                "Skip columns": this_summary["Skip columns"],
                "Value column summaries": this_summary["Value columns"],
                "Categorical column summaries": this_summary["Categorical columns"],
                "Categorical counts": this_summary["Categorical counts"],
            },
        }

    def merge_all_info(self) -> "TabularSummary":
        """Create a TabularSummary containing the overall dataset summary.

        Returns:
            TabularSummary - the summary object for column values.

        """
        all_sum = TabularSummary(value_cols=self.op.value_columns, skip_cols=self.op.skip_columns, name="Dataset")
        for counts in self.summary_dict.values():
            all_sum.update_summary(counts)
        return all_sum

    def _get_result_string(self, name, summary, individual=False) -> str:
        """Return a formatted string with the summary for the indicated name.

        Parameters:
            name (str):  Identifier (usually the filename) of the individual file.
            summary (dict): The dictionary of the summary results indexed by name.
            individual (bool): Whether this is for an individual file summary.

        Returns:
            str: The results in a printable format ready to be saved to a text file.

        Notes:
            This calls _get_dataset_string to get the overall summary string and
            _get_individual_string to get an individual summary string.

        """

        if name == "Dataset":
            sum_list = [
                f"Dataset: Total events={summary.get('Total events', 0)} " f"Total files={summary.get('Total files', 0)}"
            ]
        else:
            sum_list = [f"Total events={summary.get('Total events', 0)}"]
        sum_list = sum_list + self._get_detail_list(summary, indent=BaseSummary.DISPLAY_INDENT)
        return "\n".join(sum_list)

    def _get_individual_string(self, result, indent=BaseSummary.DISPLAY_INDENT) -> str:
        """Return a formatted string with the summary for an individual file.

        Parameters:
            result (dict): The dictionary of the summary results indexed by name.
            indent (str): A string containing spaces used for indentation (usually 3 spaces).

        Returns:
            str: The results in a printable format ready to be saved to a text file.

        Notes:
            This calls _get_categorical_string to get the categorical part of the summary,
            and _get_value_string to get the value column part of the summary.

        """

        return "\n".join(
            [
                f"Summary for {result['Name']}",
                f"  Total events: {result.get('Total events', 0)}",
                f"  Total files: {result.get('Total files', 0)}",
                f"  Value columns: {result['Specifics']['Value columns']}",
                f"  Skip columns: {result['Specifics']['Skip columns']}",
                self._get_categorical_string(result),
                self._get_value_string(result["Specifics"]["Value column summaries"]),
            ]
        )

    def _format_categorical_lists(self, specifics) -> list:
        """Format the categorical column summaries for display.

        Parameters:
            specifics (dict): The specifics dictionary from the summary.

        Returns:
            list: A list of formatted strings for the categorical column summaries.

        """
        cat_dict = specifics.get("Categorical column summaries", {})
        if not cat_dict:
            return []
        count_dict = specifics["Categorical counts"]
        formatted_list = ["Categorical column values[Events, Files]:"]
        sorted_tuples = sorted(cat_dict.items(), key=lambda x: x[0])
        for entry in sorted_tuples:
            formatted_list = formatted_list + self._get_categorical_col(entry, count_dict, offset="", indent="   ")
        return formatted_list

    def _get_categorical_string(self, summary, offset="", indent="   "):
        """Return  a string with the summary for a particular categorical dictionary.

        Parameters:
            summary (dict): Dictionary of summary information for a particular tabular file.
            offset (str): String of blanks used as offset for every item
            indent (str):  String of blanks used as the additional amount to indent an item's for readability.

        Returns:
            str: Formatted string suitable for saving in a file or printing.

        """
        specifics = summary.get("Specifics", {})
        cat_dict = specifics.get("Categorical column summaries", {})
        if not cat_dict:
            return ""
        count_dict = specifics.get("Categorical counts", {})
        sum_list = [f"{offset}{indent}Categorical column values[Events, Files]:"]
        sorted_tuples = sorted(cat_dict.items(), key=lambda x: x[0])
        for entry in sorted_tuples:
            sum_list = sum_list + self._get_categorical_col(entry, count_dict, offset="", indent="   ")
        return "\n".join(sum_list)

    def _get_detail_list(self, result, indent=BaseSummary.DISPLAY_INDENT):
        """Return  a list of strings with the details

        Parameters:
            result (dict): Dictionary of merged summary information.
            indent (str):  String of blanks used as the amount to indent for readability.

        Returns:
            list: list of formatted strings suitable for saving in a file or printing.

        """
        sum_list = []
        specifics = result["Specifics"]
        cat_string = self._get_categorical_string(specifics, offset="", indent=indent)
        if cat_string:
            sum_list.append(cat_string)
        val_dict = specifics.get("Value column summaries", {})
        if val_dict:
            sum_list.append(ColumnValueSummary._get_value_string(val_dict, offset="", indent=indent))
        return sum_list

    def _get_categorical_col(self, entry, count_dict, offset="", indent="   "):
        """Return  a string with the summary for a particular categorical column.

        Parameters:
            entry(tuple): (Name of the column, summary dict for that column)
            count_dict (dict): Count of the total number of unique values indexed by the name
            offset(str): String of blanks used as offset for all items
            indent (str):  String of blanks used as the additional amount to indent for this item's readability.

        Returns:
            list: Formatted strings, each corresponding to a line in the output.
        """
        num_unique = count_dict[entry[0]]
        num_disp = min(self.op.max_categorical, num_unique)
        col_list = [f"{offset}{indent * 2}{entry[0]}: {num_unique} unique values " f"(displaying top {num_disp} values)"]
        # Create and partition the list of individual entries
        value_list = [f"{item[0]}{str(item[1])}" for item in entry[1].items()]
        value_list = value_list[:num_disp]
        part_list = ColumnValueSummary.partition_list(value_list, self.op.values_per_line)
        return col_list + [f"{offset}{indent * 3}{ColumnValueSummary.get_list_str(item)}" for item in part_list]

    @staticmethod
    def get_list_str(lst) -> str:
        """Return a str version of a list with items separated by a blank.

        Returns:
            str:  String version of list.

        """
        return f"{' '.join(str(item) for item in lst)}"

    @staticmethod
    def partition_list(lst, n) -> list:
        """Partition a list into lists of n items.

        Parameters:
            lst (list): List to be partitioned.
            n (int):  Number of items in each sublist.

        Returns:
            list:  list of lists of n elements, the last might have fewer.

        """
        return [lst[i : i + n] for i in range(0, len(lst), n)]

    @staticmethod
    def _get_value_string(val_dict, offset="", indent="") -> str:
        sum_list = [f"{offset}{indent}Value columns[Events, Files]:"]
        for col_name, val_counts in val_dict.items():
            sum_list.append(f"{offset}{indent*2}{col_name}{str(val_counts)}")
        return "\n".join(sum_list)

    @staticmethod
    def sort_dict(count_dict, reverse=False):
        sorted_tuples = sorted(count_dict.items(), key=lambda x: x[1][0], reverse=reverse)
        return len(sorted_tuples), sorted_tuples



"""Abstract base class for the contents of summary operations."""

import os
from abc import ABC, abstractmethod
import json
from hed.tools.util import io_util


class BaseSummary(ABC):
    """Abstract base class for summary contents. Should not be instantiated.

    Parameters:
        sum_op (BaseOp):  Operation corresponding to this summary.

    """

    DISPLAY_INDENT = "   "
    INDIVIDUAL_SUMMARIES_PATH = "individual_summaries"

    def __init__(self, sum_op):
        self.op = sum_op
        self.summary_dict = {}

    def get_summary_details(self, include_individual=True) -> dict:
        """Return a dictionary with the details for individual files and the overall dataset.

        Parameters:
            include_individual (bool):  If True, summaries for individual files are included.

        Returns:
            dict: A dictionary with 'Dataset' and 'Individual files' keys.

        Notes:
            - The 'Dataset' value is either a string or a dictionary with the overall summary.
            - The 'Individual files' value is dictionary whose keys are file names and values are
                   their corresponding summaries.

        Users are expected to provide merge_all_info and get_details_dict functions to support this.

        """
        merged_counts = self.merge_all_info()
        if merged_counts:
            details = self.get_details_dict(merged_counts)
        else:
            details = "Overall summary unavailable"

        summary_details = {"Dataset": details, "Individual files": {}}
        if include_individual:
            for name, count in self.summary_dict.items():
                summary_details["Individual files"][name] = self.get_details_dict(count)
        return summary_details

    def get_summary(self, individual_summaries="separate"):
        """Return a summary dictionary with the information.

        Parameters:
            individual_summaries (str): "separate", "consolidated", or "none"

        Returns:
            dict: Dictionary with "Dataset" and "Individual files" keys.

        Notes: The individual_summaries value is processed as follows:
           -  "separate" individual summaries are to be in separate files.
           -  "consolidated" means that the individual summaries are in same file as overall summary.
           -  "none" means that only the overall summary is produced.

        """
        include_individual = individual_summaries == "separate" or individual_summaries == "consolidated"
        summary_details = self.get_summary_details(include_individual=include_individual)
        dataset_summary = {
            "Summary name": self.op.summary_name,
            "Summary type": self.op.SUMMARY_TYPE,
            "Summary filename": self.op.summary_filename,
            "Overall summary": summary_details["Dataset"],
        }
        summary = {"Dataset": dataset_summary, "Individual files": {}}
        if summary_details["Individual files"]:
            summary["Individual files"] = self.get_individual(
                summary_details["Individual files"], separately=individual_summaries == "separate"
            )
        return summary

    def get_individual(self, summary_details, separately=True):
        """Return a dictionary of the individual file summaries.

        Parameters:
            summary_details (dict): Dictionary of the individual file summaries.
            separately (bool): If True (the default), each individual summary has a header for separate output.
        """
        individual_dict = {}
        for name, name_summary in summary_details.items():
            if separately:
                individual_dict[name] = {
                    "Summary name": self.op.summary_name,
                    "summary type": self.op.SUMMARY_TYPE,
                    "Summary filename": self.op.summary_filename,
                    "File summary": name_summary,
                }
            else:
                individual_dict[name] = name_summary
        return individual_dict

    def get_text_summary_details(self, include_individual=True) -> dict:
        """Return a text summary of the information represented by this summary.

        Parameters:
            include_individual (bool): If True (the default), individual summaries are in "Individual files".

        Returns:
            dict: Dictionary with "Dataset" and "Individual files" keys.

        """
        result = self.get_summary_details(include_individual=include_individual)
        summary_details = {"Dataset": self._get_result_string("Dataset", result.get("Dataset", "")), "Individual files": {}}
        if include_individual:
            for name, individual_result in result.get("Individual files", {}).items():
                summary_details["Individual files"][name] = self._get_result_string(name, individual_result)
        return summary_details

    def get_text_summary(self, individual_summaries="separate") -> dict:
        """Return a complete text summary by assembling the individual pieces.

        Parameters:
            individual_summaries(str):  One of the values "separate", "consolidated", or "none".

        Returns:
            dict: Complete text summary.

        Notes: The options are:
            - "none":  Just has "Dataset" key.
            - "consolidated"  Has "Dataset" and "Individual files" keys with the values of each is a string.
            - "separate" Has "Dataset" and "Individual files" keys. The values of "Individual files" is a dict.

        """
        include_individual = individual_summaries == "separate" or individual_summaries == "consolidated"
        summary_details = self.get_text_summary_details(include_individual=include_individual)
        summary = {
            "Dataset": f"Summary name: {self.op.summary_name}\n"
            + f"Summary type: {self.op.SUMMARY_TYPE}\n"
            + f"Summary filename: {self.op.summary_filename}\n\n"
            + f"Overall summary:\n{summary_details['Dataset']}"
        }
        if individual_summaries == "separate":
            summary["Individual files"] = {}
            for name, name_summary in summary_details["Individual files"].items():
                summary["Individual files"][name] = (
                    f"Summary name: {self.op.summary_name}\n"
                    + f"Summary type: {self.op.SUMMARY_TYPE}\n"
                    + f"Summary filename: {self.op.summary_filename}\n\n"
                    + f"Summary for {name}:\n{name_summary}"
                )
        elif include_individual:
            ind_list = []
            for name, name_summary in summary_details["Individual files"].items():
                ind_list.append(f"{name}:\n{name_summary}\n")
            ind_str = "\n\n".join(ind_list)
            summary["Dataset"] = summary["Dataset"] + f"\n\nIndividual files:\n\n{ind_str}"

        return summary

    def save(self, save_dir, file_formats=None, individual_summaries="separate", task_name=""):
        """Save the summaries using the format indicated.

        Parameters:
            save_dir (str):  Name of the directory to save the summaries in.
            file_formats (list or None):  List of file formats to use for saving. If None, defaults to ['.txt'].
            individual_summaries (str):  Save one file or multiple files based on setting.
            task_name (str): If this summary corresponds to files from a task, the task_name is used in filename.

        """
        if file_formats is None:
            file_formats = [".txt"]
        for file_format in file_formats:
            if file_format == ".txt":
                summary = self.get_text_summary(individual_summaries=individual_summaries)
            elif file_format == ".json":
                summary = self.get_summary(individual_summaries=individual_summaries)
            else:
                continue
            self._save_summary_files(save_dir, file_format, summary, individual_summaries, task_name=task_name)

            self.save_visualizations(
                save_dir, file_formats=file_formats, individual_summaries=individual_summaries, task_name=task_name
            )

    def save_visualizations(self, save_dir, file_formats=None, individual_summaries="separate", task_name=""):
        """Save summary visualizations, if any, using the format indicated.

        Parameters:
            save_dir (str):  Name of the directory to save the summaries in.
            file_formats (list or None):  List of file formats to use for saving. If None, defaults to ['.svg'].
            individual_summaries (str):  Save one file or multiple files based on setting.
            task_name (str): If this summary corresponds to files from a task, the task_name is used in filename.

        """
        if file_formats is None:
            file_formats = [".svg"]
        pass

    def _save_summary_files(self, save_dir, file_format, summary, individual_summaries, task_name=""):
        """Save the files in the appropriate format.

        Parameters:
            save_dir (str): Path to the directory in which the summaries will be saved.
            file_format (str): string representing the extension (including .), '.txt' or '.json'.
            summary (dictionary): Dictionary of summaries (has "Dataset" and "Individual files" keys).
            individual_summaries (str): "consolidated", "individual", or "none".
            task_name (str): Name of task to be included in file name if multiple tasks.

        """
        if self.op.append_timecode:
            time_stamp = "_" + io_util.get_timestamp()
        else:
            time_stamp = ""
        if task_name:
            task_name = "_" + task_name
        this_save = os.path.join(save_dir, self.op.summary_name + "/")
        os.makedirs(os.path.realpath(this_save), exist_ok=True)
        filename = os.path.realpath(os.path.join(this_save, self.op.summary_filename + task_name + time_stamp + file_format))
        individual = summary.get("Individual files", {})
        if individual_summaries == "none" or not individual:
            self.dump_summary(filename, summary["Dataset"])
            return
        if individual_summaries == "consolidated":
            self.dump_summary(filename, summary)
            return
        self.dump_summary(filename, summary["Dataset"])
        individual_dir = os.path.join(this_save, self.INDIVIDUAL_SUMMARIES_PATH + "/")
        os.makedirs(os.path.realpath(individual_dir), exist_ok=True)
        for name, sum_str in individual.items():
            filename = self._get_summary_filepath(individual_dir, name, task_name, time_stamp, file_format)
            self.dump_summary(filename, sum_str)

    def _get_summary_filepath(self, individual_dir, name, task_name, time_stamp, file_format):
        """Return the filepath for the summary including the timestamp

        Parameters:
            individual_dir (str):  path of the directory in which the summary should be stored.
            name (str): Path of the original file from which the summary was extracted.
            task_name (str): Task name if separate summaries for different tasks or the empty string if not separated.
            time_stamp (str):  Formatted date-time string to be included in the filename of the summary.

        Returns:
            str: Full path name of the summary.

        """
        this_name = os.path.basename(name)
        this_name = os.path.splitext(this_name)[0]
        count = 1
        match = True
        filename = None
        while match:
            filename = f"{self.op.summary_filename}_{this_name}{task_name}_{count}{time_stamp}{file_format}"
            filename = os.path.realpath(os.path.join(individual_dir, filename))
            if not os.path.isfile(filename):
                break
            count = count + 1
        return filename

    def _get_result_string(self, name, result, indent=DISPLAY_INDENT):
        """Return a formatted string with the summary for the indicated name.

        Parameters:
            name (str):  Identifier (usually the filename) of the individual file.
            result (dict): The dictionary of the summary results indexed by name.
            indent (str): A string containing spaces used for indentation (usually 3 spaces).

        Returns:
            str: The results in a printable format ready to be saved to a text file.

        Notes:
            This file should be overridden by each summary.

        """
        return f"\n{name}\n{indent}{str(result)}"

    @staticmethod
    def dump_summary(filename, summary):
        with open(filename, "w") as text_file:
            if not isinstance(summary, str):
                summary = json.dumps(summary, indent=4)
            text_file.write(summary)

    @abstractmethod
    def get_details_dict(self, summary_info):
        """Return the summary-specific information.

        Parameters:
            summary_info (object):  Summary to return info from.

        Returns:
            dict: dictionary with the results.

        Notes:
            Abstract method be implemented by each individual summary.

        Notes:
            The expected return value is a dictionary of the form:

               {"Name": "", "Total events": 0, "Total files": 0, "Files": [], "Specifics": {}}"

        """
        raise NotImplementedError

    @abstractmethod
    def merge_all_info(self):
        """Return merged information.

        Returns:
           object:  Consolidated summary of information.

        Notes:
            Abstract method be implemented by each individual summary.

        """
        raise NotImplementedError

    @abstractmethod
    def update_summary(self, summary_dict):
        """Method to update summary for a given tabular input.

        Parameters:
            summary_dict (dict)  A summary specific dictionary with the update information.

        """
        raise NotImplementedError



"""Merge consecutive rows of a columnar file with same column value."""

import pandas as pd
from remodeler.operations.base_op import BaseOp


class MergeConsecutiveOp(BaseOp):
    """Merge consecutive rows of a columnar file with same column value.

    Required remodeling parameters:
        - **column_name** (*str*): name of column whose consecutive values are to be compared (the merge column).
        - **event_code** (*str* or *int* or *float*): the particular value in the match column to be merged.
        - **set_durations** (*bool*): If true, set the duration of the merged event to the extent of the merged events.
        - **ignore_missing** (*bool*):  If true, missing match_columns are ignored.

    Optional remodeling parameters:
        - **match_columns** (*list*):  A list of columns whose values have to be matched for two events to be the same.

    Notes:
          This operation is meant for time-based tabular files that have an onset column.

    """

    NAME = "merge_consecutive"

    PARAMS = {
        "type": "object",
        "properties": {
            "column_name": {
                "type": "string",
                "description": "The name of the column to check for repeated consecutive codes.",
            },
            "event_code": {"type": ["string", "number"], "description": "The event code to match for duplicates."},
            "match_columns": {
                "type": "array",
                "description": "List of columns whose values must also match to be considered a repeat.",
                "items": {"type": "string"},
            },
            "set_durations": {
                "type": "boolean",
                "description": "If true, then the duration should be computed based on start of first to end of last.",
            },
            "ignore_missing": {"type": "boolean", "description": "If true, missing match columns are ignored."},
        },
        "required": ["column_name", "event_code", "set_durations", "ignore_missing"],
        "additionalProperties": False,
    }

    def __init__(self, parameters):
        """Constructor for the merge consecutive operation.

        Parameters:
            parameters (dict): Actual values of the parameters for the operation.

        """
        super().__init__(parameters)
        self.column_name = parameters["column_name"]
        self.event_code = parameters["event_code"]
        self.set_durations = parameters["set_durations"]
        self.ignore_missing = parameters["ignore_missing"]
        self.match_columns = parameters.get("match_columns", None)

    def do_op(self, dispatcher, df, name, sidecar=None) -> pd.DataFrame:
        """Merge consecutive rows with the same column value.

        Parameters:
            dispatcher (Dispatcher): Manages the operation I/O.
            df (DataFrame): The DataFrame to be remodeled.
            name (str): Unique identifier for the dataframe -- often the original file path.
            sidecar (Sidecar or file-like): Not needed for this operation.

        Returns:
            Dataframe: A new dataframe after processing.

        :raises ValueError:
            - If dataframe does not have the anchor column and ignore_missing is False.
            - If a match column is missing and ignore_missing is False.
            - If the durations were to be set and the dataframe did not have an onset column.
            - If the durations were to be set and the dataframe did not have a duration column.

        """

        if not self.ignore_missing and self.column_name not in df.columns:
            raise ValueError(
                "ColumnMissing",
                f"{name}: {self.column_name} is not in data columns [{str(df.columns)}] "
                f"and missing columns are not ignored",
            )
        if self.set_durations and "onset" not in df.columns:
            raise ValueError("MissingOnsetColumn", f"{name}: Data must have an onset column in order to set durations")
        if self.set_durations and "duration" not in df.columns:
            raise ValueError("MissingDurationColumn", f"{name}: Data must have a duration column in order to set durations")
        missing = set(self.match_columns).difference(set(df.columns))
        if self.match_columns and not self.ignore_missing and missing:
            raise ValueError(
                "MissingMatchColumns",
                f"{name}: {str(missing)} columns are unmatched by data columns" f"[{str(df.columns)}] and not ignored",
            )
        match_columns = list(set(self.match_columns).intersection(set(df.columns)))

        df_new = df.copy()
        code_mask = df_new[self.column_name] == self.event_code
        if not code_mask.any():
            return df_new
        match_columns.append(self.column_name)
        match_df = df_new.loc[:, match_columns]
        remove_groups = self._get_remove_groups(match_df, code_mask)
        if self.set_durations and max(remove_groups) > 0:
            self._update_durations(df_new, remove_groups)
        keep_mask = [remove_group == 0 for remove_group in remove_groups]
        df_new = df_new.loc[keep_mask, :].reset_index(drop=True)
        return df_new

    @staticmethod
    def _get_remove_groups(match_df, code_mask):
        """Return a list of same length as match_df with group numbers of consecutive items.

        Parameters:
            match_df (DataFrame): DataFrame containing columns to be matched.
            code_mask (DataSeries):  Same length as match_df with the names.

        Returns:
            list:  Group numbers set (starting at 1).

        # TODO: Handle round off in rows for comparison.
        """
        in_group = False
        remove_groups = [0] * len(match_df)
        group_count = 0
        for index, row in match_df.iterrows():
            if not code_mask.iloc[index]:
                in_group = False
                continue
            elif not in_group:
                in_group = True
                group_count += 1
                continue
            if in_group and row.equals(match_df.loc[index - 1, :]):
                remove_groups[index] = group_count
            else:
                group_count += 1
        return remove_groups

    @staticmethod
    def _update_durations(df_new, remove_groups):
        """Update the durations for the columns based on merged columns.

        Parameters:
            df_new (DataFrame): Tabular data to merge.
            remove_groups (list): List of names of columns to remove.

        """
        remove_df = pd.DataFrame(remove_groups, columns=["remove"])
        max_groups = max(remove_groups)
        for index in range(max_groups):
            df_group = df_new.loc[remove_df["remove"] == index + 1, ["onset", "duration"]]
            max_group = df_group.sum(axis=1, skipna=True).max()
            anchor = df_group.index[0] - 1
            max_anchor = df_new.loc[anchor, ["onset", "duration"]].sum(skipna=True).max()
            df_new.loc[anchor, "duration"] = max(max_group, max_anchor) - df_new.loc[anchor, "onset"]

    @staticmethod
    def validate_input_data(parameters):
        """Verify that the column name is not in match columns.

        Parameters:
            parameters (dict): Dictionary of parameters of actual implementation.

        """
        match_columns = parameters.get("match_columns", None)
        name = parameters.get("column_name", None)
        if match_columns and name in match_columns:
            return [f"column_name `{name}` cannot be a match_column."]
        return []



"""Summarize the HED tags in collection of tabular files."""

import numpy as np
import pandas as pd
from hed.models.tabular_input import TabularInput
from hed.tools.analysis.hed_tag_counts import HedTagCounts
from hed.tools.analysis.event_manager import EventManager
from hed.tools.analysis.hed_tag_manager import HedTagManager
from remodeler.operations.base_op import BaseOp
from remodeler.operations.base_summary import BaseSummary


class SummarizeHedTagsOp(BaseOp):
    """Summarize the HED tags in collection of tabular files.

    Required remodeling parameters:
        - **summary_name** (*str*): The name of the summary.
        - **summary_filename** (*str*): Base filename of the summary.
        - **tags** (*dict*): Specifies how to organize the tag output.

    Optional remodeling parameters:
       - **append_timecode** (*bool*): If True, the timecode is appended to the base filename when summary is saved.
       - **include_context** (*bool*): If True, context of events is included in summary.
       - **remove_types** (*list*): A list of type tags such as Condition-variable or Task to exclude from summary.
       - **replace_defs** (*bool*): If True, the def tag is replaced by the contents of the definitions.
    The purpose of this op is to produce a summary of the occurrences of HED tags organized in a specified manner.

    Notes: The tags template is a dictionary whose keys are the organization titles (not necessarily tags) for the
    output and whose values are the tags, which if they or their children appear, they will be listed under that
    title.

    """

    NAME = "summarize_hed_tags"

    PARAMS = {
        "type": "object",
        "properties": {
            "summary_name": {"type": "string", "description": "Name to use for the summary in titles."},
            "summary_filename": {"type": "string", "description": "Name to use for the summary file name base."},
            "tags": {
                "type": "object",
                "description": "A dictionary with the template for how output of tags should be organized.",
                "patternProperties": {
                    ".*": {"type": "array", "items": {"type": "string"}, "minItems": 1, "uniqueItems": True},
                    "minProperties": 1,
                    "additionalProperties": False,
                },
            },
            "append_timecode": {
                "type": "boolean",
                "description": "If true, the timecode is appended to the base filename so each run has a unique name.",
            },
            "include_context": {
                "type": "boolean",
                "description": "If true, tags for events that unfold over time are counted at each intermediate time.",
            },
            "remove_types": {
                "type": "array",
                "description": "A list of special tags such as Condition-variable whose influence is to be removed.",
                "items": {"type": "string"},
                "minItems": 1,
                "uniqueItems": True,
            },
            "replace_defs": {
                "type": "boolean",
                "description": "If true, then the Def tags are replaced with actual definitions for the count.",
            },
        },
        "required": ["summary_name", "summary_filename", "tags"],
        "additionalProperties": False,
    }

    SUMMARY_TYPE = "hed_tag_summary"

    def __init__(self, parameters):
        """Constructor for the summarize_hed_tags operation.

        Parameters:
            parameters (dict): Dictionary with the parameter values for required and optional parameters.

        """
        super().__init__(parameters)
        self.summary_name = parameters["summary_name"]
        self.summary_filename = parameters["summary_filename"]
        self.tags = parameters["tags"]
        self.append_timecode = parameters.get("append_timecode", False)
        self.include_context = parameters.get("include_context", True)
        self.replace_defs = parameters.get("replace_defs", True)
        self.remove_types = parameters.get("remove_types", [])

    def do_op(self, dispatcher, df, name, sidecar=None) -> pd.DataFrame:
        """Summarize the HED tags present in the dataset.

        Parameters:
            dispatcher (Dispatcher): Manages the operation I/O.
            df (DataFrame): The DataFrame to be remodeled.
            name (str): Unique identifier for the dataframe -- often the original file path.
            sidecar (Sidecar or file-like):  Only needed for HED operations.

        Returns:
            DataFrame: A copy of df.

        Side effect:
            Updates the context.

        """
        df_new = df.copy()
        summary = dispatcher.summary_dicts.get(self.summary_name, None)
        if not summary:
            summary = HedTagSummary(self)
            dispatcher.summary_dicts[self.summary_name] = summary
        summary.update_summary(
            {"df": dispatcher.post_proc_data(df_new), "name": name, "schema": dispatcher.hed_schema, "sidecar": sidecar}
        )
        return df_new

    @staticmethod
    def validate_input_data(parameters):
        """Additional validation required of operation parameters not performed by JSON schema validator."""
        return []


class HedTagSummary(BaseSummary):
    """Manager of the HED tag summaries."""

    def __init__(self, sum_op):
        """Constructor for HED tag summary manager.

        Parameters:
            sum_op (SummarizeHedTagsOp): Operation associated with this summary.

        """

        super().__init__(sum_op)
        self.sum_op = sum_op

    def update_summary(self, new_info):
        """Update the summary for a given tabular input file.

        Parameters:
            new_info (dict):  A dictionary with the parameters needed to update a summary.

        Notes:
            - The summary needs a "name" str, a "schema", a "df, and a "Sidecar".

        """
        counts = HedTagCounts(new_info["name"], total_events=len(new_info["df"]))
        input_data = TabularInput(new_info["df"], sidecar=new_info["sidecar"], name=new_info["name"])
        tag_man = HedTagManager(EventManager(input_data, new_info["schema"]), remove_types=self.sum_op.remove_types)
        hed_objs = tag_man.get_hed_objs(include_context=self.sum_op.include_context, replace_defs=self.sum_op.replace_defs)
        for hed in hed_objs:
            counts.update_tag_counts(hed, new_info["name"])
        self.summary_dict[new_info["name"]] = counts

    def get_details_dict(self, tag_counts) -> dict:
        """Return the summary-specific information in a dictionary.

        Parameters:
            tag_counts (HedTagCounts):  Contains the counts of tags in the dataset.

        Returns:
            dict: dictionary with the summary results.

        """
        template, unmatched = tag_counts.organize_tags(self.sum_op.tags)
        details = {}
        for key, key_list in self.sum_op.tags.items():
            details[key] = self._get_details(key_list, template, verbose=True)
        leftovers = [value.get_info(verbose=True) for value in unmatched]
        return {
            "Name": tag_counts.name,
            "Total events": tag_counts.total_events,
            "Total files": len(tag_counts.files.keys()),
            "Files": list(tag_counts.files.keys()),
            "Specifics": {"Main tags": details, "Other tags": leftovers},
        }

    def _get_result_string(self, name, result, indent=BaseSummary.DISPLAY_INDENT):
        """Return a formatted string with the summary for the indicated name.

        Parameters:
            name (str):  Identifier (usually the filename) of the individual file.
            result (dict): The dictionary of the summary results indexed by name.
            indent (str): A string containing spaces used for indentation (usually 3 spaces).

        Returns:
            str: The results in a printable format ready to be saved to a text file.

        Notes:
            This calls _get_dataset_string to get the overall summary string and
            _get_individual_string to get an individual summary string.

        """
        if name == "Dataset":
            return self._get_dataset_string(result, indent=indent)
        return self._get_individual_string(result, indent=indent)

    def merge_all_info(self) -> "HedTagCounts":
        """Create a HedTagCounts containing the overall dataset HED tag  summary.

        Returns:
            HedTagCounts: The overall dataset summary object for HED tag counts.

        """

        all_counts = HedTagCounts("Dataset")
        for _key, counts in self.summary_dict.items():
            all_counts.merge_tag_dicts(counts.tag_dict)
            for file_name in counts.files.keys():
                all_counts.files[file_name] = ""
            all_counts.total_events = all_counts.total_events + counts.total_events
        return all_counts

    @staticmethod
    def summary_to_dict(specifics, transform=np.log10, scale_adjustment=7) -> dict:
        """Convert a HedTagSummary json specifics dict into the word cloud input format.

        Parameters:
            specifics (dict): Dictionary with keys "Main tags" and "Other tags".
            transform (func): The function to transform the number of found tags.
                             Default log10
            scale_adjustment (int): Value added after transform.

        Returns:
            dict: A dict of the words and their occurrence count.

        Raises:
            KeyError: A malformed dictionary was passed.

        """
        if transform is None:

            def transform(x):
                return x

        word_dict = {}
        tag_dict = specifics.get("Main tags", {})
        for tag, tag_sub_list in tag_dict.items():
            if tag == "Exclude tags":
                continue
            for tag_sub_dict in tag_sub_list:
                word_dict[tag_sub_dict["tag"]] = transform(tag_sub_dict["events"]) + scale_adjustment
        other_dict = specifics.get("Other tags", [])
        for tag_sub_list in other_dict:
            word_dict[tag_sub_list["tag"]] = transform(tag_sub_list["events"]) + scale_adjustment
        return word_dict

    @staticmethod
    def _get_dataset_string(result, indent=BaseSummary.DISPLAY_INDENT):
        """Return  a string with the overall summary for all the tabular files.

        Parameters:
            result (dict): Dictionary of merged summary information.
            indent (str):  String of blanks used as the amount to indent for readability.

        Returns:
            str: Formatted string suitable for saving in a file or printing.

        """
        sum_list = [f"Dataset: Total events={result.get('Total events', 0)} " f"Total files={len(result.get('Files', []))}"]
        sum_list = sum_list + HedTagSummary._get_tag_list(result, indent=indent)
        return "\n".join(sum_list)

    @staticmethod
    def _get_individual_string(result, indent=BaseSummary.DISPLAY_INDENT):
        """Return  a string with the summary for an individual tabular file.

        Parameters:
            result (dict): Dictionary of summary information for a particular tabular file.
            indent (str):  String of blanks used as the amount to indent for readability.

        Returns:
            str: Formatted string suitable for saving in a file or printing.

        """
        sum_list = [f"Total events={result.get('Total events', 0)}"]
        sum_list = sum_list + HedTagSummary._get_tag_list(result, indent=indent)
        return "\n".join(sum_list)

    @staticmethod
    def _tag_details(tags):
        """Return a list of strings with the tag details.

        Parameters:
            tags (list): List of tags to summarize.

        Returns:
            list: Each entry has the summary details for a tag.

        """
        tag_list = []
        for tag in tags:
            tag_list.append(f"{tag['tag']}[{tag['events']},{len(tag['files'])}]")
        return tag_list

    @staticmethod
    def _get_tag_list(result, indent=BaseSummary.DISPLAY_INDENT):
        """Return a list lines to be output to summarize the tags as organized in the result.

        Parameters:
            result (dict):  Dictionary with the results organized under key "Specifics".
            indent (str):  Spaces to indent each line.

        Returns:
            list:  Each entry is a string representing a line to be printed.

        """
        tag_info = result["Specifics"]
        sum_list = [f"\n{indent}Main tags[events,files]:"]
        for category, tags in tag_info["Main tags"].items():
            sum_list.append(f"{indent}{indent}{category}:")
            if tags:
                sum_list.append(f"{indent}{indent}{indent}{' '.join(HedTagSummary._tag_details(tags))}")
        if tag_info["Other tags"]:
            sum_list.append(f"{indent}Other tags[events,files]:")
            sum_list.append(f"{indent}{indent}{' '.join(HedTagSummary._tag_details(tag_info['Other tags']))}")
        return sum_list

    @staticmethod
    def _get_details(key_list, template, verbose=False):
        """Organized a tag information from a list based on the template.

        Parameters:
            key_list (list): List of information to be organized based on the template.
            template (dict): An input template derived from the input parameters.
            verbose (bool): If False (the default) output minimal information about the summary.

        """
        key_details = []
        for item in key_list:
            for tag_cnt in template[item.casefold()]:
                key_details.append(tag_cnt.get_info(verbose=verbose))
        return key_details



"""Summarize the type_defs in the dataset."""

import pandas as pd
from hed.models.tabular_input import TabularInput
from remodeler.operations.base_op import BaseOp
from remodeler.operations.base_summary import BaseSummary
from hed.models.def_expand_gather import DefExpandGatherer


class SummarizeDefinitionsOp(BaseOp):
    """Summarize the definitions used in the dataset based on Def and Def-expand.

    Required remodeling parameters:
        - **summary_name** (*str*): The name of the summary.
        - **summary_filename** (*str*): Base filename of the summary.

    Optional remodeling parameters:
         - **append_timecode** (*bool*): If False (default), the timecode is not appended to the summary filename.

    The purpose is to produce a summary of the definitions used in a dataset.

    """

    NAME = "summarize_definitions"

    PARAMS = {
        "type": "object",
        "properties": {
            "summary_name": {"type": "string", "description": "Name to use for the summary in titles."},
            "summary_filename": {"type": "string", "description": "Name to use for the summary file name base."},
            "append_timecode": {
                "type": "boolean",
                "description": "If true, the timecode is appended to the base filename so each run has a unique name.",
            },
        },
        "required": ["summary_name", "summary_filename"],
        "additionalProperties": False,
    }

    SUMMARY_TYPE = "type_defs"

    def __init__(self, parameters):
        """Constructor for the summary of definitions used in the dataset.

        Parameters:
            parameters (dict): Dictionary with the parameter values for required and optional parameters.

        """
        super().__init__(parameters)
        self.summary_name = parameters["summary_name"]
        self.summary_filename = parameters["summary_filename"]
        self.append_timecode = parameters.get("append_timecode", False)

    def do_op(self, dispatcher, df, name, sidecar=None) -> pd.DataFrame:
        """Create summaries of definitions.

        Parameters:
            dispatcher (Dispatcher): Manages the operation I/O.
            df (DataFrame): The DataFrame to be remodeled.
            name (str):  Unique identifier for the dataframe -- often the original file path.
            sidecar (Sidecar or file-like): Only needed for HED operations.

        Returns:
            DataFrame: a copy of df

        Side effect:
            Updates the relevant summary.

        """
        df_new = df.copy()
        summary = dispatcher.summary_dicts.setdefault(self.summary_name, DefinitionSummary(self, dispatcher.hed_schema))
        summary.update_summary(
            {"df": dispatcher.post_proc_data(df_new), "name": name, "sidecar": sidecar, "schema": dispatcher.hed_schema}
        )
        return df_new

    @staticmethod
    def validate_input_data(parameters):
        """Additional validation required of operation parameters not performed by JSON schema validator."""
        return []


class DefinitionSummary(BaseSummary):
    """Manager for summaries of the definitions used in a dataset."""

    def __init__(self, sum_op, hed_schema, known_defs=None):
        """Constructor for the summary of definitions.

        Parameters:
            sum_op (SummarizeDefinitionsOp): Summary operation class for gathering definitions.
            hed_schema (HedSchema or HedSchemaGroup):  Schema used for the dataset.
            known_defs (str or list or DefinitionDict): Definitions already known to be used.


        """
        super().__init__(sum_op)
        self.def_gatherer = DefExpandGatherer(hed_schema, known_defs=known_defs)

    def update_summary(self, new_info):
        """Update the summary for a given tabular input file.

        Parameters:
            new_info (dict):  A dictionary with the parameters needed to update a summary.

        Notes:
            - The summary needs a "name" str, a "schema" and a "Sidecar".

        """
        data_input = TabularInput(new_info["df"], sidecar=new_info["sidecar"], name=new_info["name"])
        series, def_dict = data_input.series_a, data_input.get_def_dict(new_info["schema"])
        self.def_gatherer.process_def_expands(series, def_dict)

    @staticmethod
    def _build_summary_dict(items_dict, title, process_func, display_description=False):
        summary_dict = {}
        items = {}
        for key, value in items_dict.items():
            if process_func:
                value = process_func(value)
            if "#" in str(value):
                key = key + "/#"
            if display_description:
                description, value = DefinitionSummary._remove_description(value)
                items[key] = {"description": description, "contents": str(value)}
            elif isinstance(value, list):
                items[key] = [str(x) for x in value]
            else:
                items[key] = str(value)
        summary_dict[title] = items
        return summary_dict

    def get_details_dict(self, def_summary) -> dict:
        """Return the summary-specific information in a dictionary.

        Parameters:
            def_summary (DefExpandGatherer):  Contains the resolved dictionaries.

        Returns:
            dict: dictionary with the summary results.

        """
        known_defs_summary = self._build_summary_dict(
            def_summary.def_dict, "Known Definitions", None, display_description=True
        )
        # ambiguous_defs_summary = self._build_summary_dict(def_gatherer.ambiguous_defs, "Ambiguous Definitions",
        #                                                   def_gatherer.get_ambiguous_group)
        # ambiguous_defs_summary = {}
        # TODO: Summary of ambiguous definitions is not implemented
        errors_summary = self._build_summary_dict(def_summary.errors, "Errors", None)

        known_defs_summary.update(errors_summary)
        return {"Name": "", "Total events": 0, "Total files": 0, "Files": [], "Specifics": known_defs_summary}
        # return known_defs_summary

    def merge_all_info(self) -> object:
        """Create an Object containing the definition summary.

        Returns:
            Object: The overall summary object for type_defs.

        """
        return self.def_gatherer

    def _get_result_string(self, name, result, indent=BaseSummary.DISPLAY_INDENT):
        """Return a formatted string with the summary for the indicated name.

        Parameters:
            name (str):  Identifier (usually the filename) of the individual file.
            result (dict): The dictionary of the summary results indexed by name.
            indent (str): A string containing spaces used for indentation (usually 3 spaces).

        Returns:
            str: The results in a printable format ready to be saved to a text file.

        Notes:
            This calls _get_dataset_string to get the overall summary string and
            _get_individual_string to get an individual summary string.

        """
        if name == "Dataset":
            return self._get_dataset_string(result, indent=indent)
        return self._get_individual_string(result, indent=indent)

    @staticmethod
    def _nested_dict_to_string(data, indent, level=1):
        """Return string summary of definitions used by recursively traversing the summary info.

        Parameters:
            data (dict):  Dictionary containing information.
            indent (str):  Spaces to indent the nested results.
            level (int):  (Default 1): Level indicator for recursive calls.

        """
        result = []
        for key, value in data.items():
            if isinstance(value, dict):
                result.append(f"{indent * level}{key}: {len(value)} items")
                result.append(DefinitionSummary._nested_dict_to_string(value, indent, level + 1))
            elif isinstance(value, list):
                result.append(f"{indent * level}{key}:")
                for item in value:
                    result.append(f"{indent * (level + 1)}{item}")
            else:
                result.append(f"{indent * level}{key}: {value}")
        return "\n".join(result)

    @staticmethod
    def _get_dataset_string(summary_dict, indent=BaseSummary.DISPLAY_INDENT):
        """Return the string representing the summary of the definitions across the dataset.

        Parameters:
            summary_dict (dict): Contains the merged summary information.
            indent (str): Spaces to indent successively levels.

        Returns:
            str:  String summary of the definitions used in the dataset.

        """
        return DefinitionSummary._nested_dict_to_string(summary_dict, indent)

    @staticmethod
    def _remove_description(def_entry):
        """Remove description from a definition entry.

        Parameters:
            def_entry (DefinitionEntry): Definition entry from which to remove its definition.

        Returns:
            tuple[str, DefinitionEntry]:
            - Description string.
            - DefinitionEntry after description has been removed.


        """
        def_group = def_entry.contents.copy()
        description = ""
        desc_tag = def_group.find_tags({"description"}, include_groups=False)
        if desc_tag:
            def_group.remove(desc_tag)
            desc_tag = desc_tag[0]
            description = desc_tag.extension

        return description, def_group

    @staticmethod
    def _get_individual_string(result, indent=BaseSummary.DISPLAY_INDENT):
        """Return  a string with the summary for an individual tabular file.

        Parameters:
            result (dict): Dictionary of summary information for a particular tabular file.
            indent (str):  String of blanks used as the amount to indent for readability.

        Returns:
            str: Formatted string suitable for saving in a file or printing.

        """
        return ""



"""Implementation in progress."""

from remodeler.operations.base_op import BaseOp


# TODO: This class is under development


class NumberGroupsOp(BaseOp):
    """Implementation in progress."""

    NAME = "number_groups"

    PARAMS = {
        "type": "object",
        "properties": {
            "number_column_name": {"type": "string"},
            "source_column": {"type": "string"},
            "start": {
                "type": "object",
                "properties": {"values": {"type": "array"}, "inclusion": {"type": "string", "enum": ["include", "exclude"]}},
                "required": ["values", "inclusion"],
                "additionalProperties": False,
            },
            "stop": {
                "type": "object",
                "properties": {"values": {"type": "array"}, "inclusion": {"type": "string", "enum": ["include", "exclude"]}},
                "required": ["values", "inclusion"],
                "additionalProperties": False,
            },
            "overwrite": {"type": "boolean"},
        },
        "required": ["number_column_name", "source_column", "start", "stop"],
        "additionalProperties": False,
    }

    def __init__(self, parameters):
        super().__init__(parameters)
        self.number_column_name = parameters["number_column_name"]
        self.source_column = parameters["source_column"]
        self.start = parameters["start"]
        self.stop = parameters["stop"]
        self.start_stop_test = {"values": list, "inclusion": str}
        self.inclusion_test = ["include", "exclude"]
        self.overwrite = parameters.get("overwrite", False)

    def do_op(self, dispatcher, df, name, sidecar=None):
        """Add numbers to groups of events in dataframe.

        Parameters:
            dispatcher (Dispatcher): Manages the operation I/O.
            df (DataFrame):  The DataFrame to be remodeled.
            name (str):  Unique identifier for the dataframe -- often the original file path.
            sidecar (Sidecar or file-like):  Only needed for HED operations.

        Returns:
            Dataframe: A new dataframe after processing.

        """
        # check if number_column_name exists and if so, check overwrite setting
        if self.number_column_name in df.columns:
            if self.overwrite is False:
                raise ValueError("ExistingNumberColumn", f"Column {self.number_column_name} already exists in event file.", "")

        # check if source_column exists
        if self.source_column not in df.columns:
            raise ValueError("MissingSourceColumn", f"Column {self.source_column} does not exist in event file {name}.", "")

        # check if all elements in value lists start and stop exist in the source_column
        missing = []
        for element in self.start["values"]:
            if element not in df[self.source_column].tolist():
                missing.append(element)
        if len(missing) > 0:
            raise ValueError(
                "MissingValue", f"Start value(s) {missing} does not exist in {self.source_column} of event file {name}"
            )

        missing = []
        for element in self.stop["values"]:
            if element not in df[self.source_column].tolist():
                missing.append(element)
        if len(missing) > 0:
            raise ValueError(
                "MissingValue", f"Start value(s) {missing} does not exist in {self.source_column} of event file {name}"
            )

        df_new = df.copy()
        return df_new

    @staticmethod
    def validate_input_data(parameters):
        """Additional validation required of operation parameters not performed by JSON schema validator."""
        return []



import os
import io
import shutil
import unittest
from unittest.mock import patch
import zipfile
from hed.errors import HedFileError
from remodeler.backup_manager import BackupManager
from hed.tools.util.io_util import get_file_list


class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        file_list = ["top_level.tsv", "sub1/sub1_events.tsv", "sub2/sub2_events.tsv", "sub2/sub2_next_events.tsv"]
        cls.file_list = file_list
        cls.extract_path = os.path.realpath(os.path.join(os.path.dirname(__file__), "./data/"))
        test_root = os.path.realpath(os.path.join(os.path.dirname(__file__), "./data/test_root"))
        cls.test_root = test_root
        cls.test_paths = [os.path.join(test_root, file) for file in file_list]
        cls.test_zip = os.path.realpath(os.path.join(os.path.dirname(__file__), "./data/test_root.zip"))

        test_root_back1 = os.path.realpath(os.path.join(os.path.dirname(__file__), "./data/test_root_back1"))
        cls.test_root_back1 = test_root_back1
        cls.test_paths_back1 = [os.path.join(test_root_back1, file) for file in file_list]
        cls.test_zip_back1 = os.path.realpath(os.path.join(os.path.dirname(__file__), "./data/test_root_back1.zip"))

        test_root_bad = os.path.realpath(os.path.join(os.path.dirname(__file__), "./data/test_root_bad"))
        cls.test_root_bad = test_root_bad
        cls.test_root_bad_backups = os.path.join(test_root_bad, BackupManager.RELATIVE_BACKUP_LOCATION)
        cls.test_paths_bad = [os.path.join(test_root_bad, file) for file in file_list]
        cls.test_zip_bad = os.path.realpath(os.path.join(os.path.dirname(__file__), "./data/test_root_bad.zip"))

    def setUp(self):
        with zipfile.ZipFile(self.test_zip, "r") as zip_ref:
            zip_ref.extractall(self.extract_path)
        with zipfile.ZipFile(self.test_zip_back1, "r") as zip_ref:
            zip_ref.extractall(self.extract_path)
        with zipfile.ZipFile(self.test_zip_bad, "r") as zip_ref:
            zip_ref.extractall(self.extract_path)

    def tearDown(self):
        shutil.rmtree(self.test_root)
        shutil.rmtree(self.test_root_back1)
        shutil.rmtree(self.test_root_bad)

    @classmethod
    def tearDownClass(cls):
        pass

    def test_constructor(self):
        back1_man = BackupManager(self.test_root_back1)
        self.assertIsInstance(back1_man, BackupManager, "constructor creates a BackupManager if no backups")
        self.assertTrue(back1_man.backups_dict)

    def test_constructor_alternative_location(self):
        alt_path = os.path.realpath(os.path.join(self.extract_path, "temp_backs"))
        back1_man = BackupManager(self.test_root_back1, backups_root=alt_path)
        self.assertIsInstance(back1_man, BackupManager, "constructor creates a BackupManager if no backups")
        self.assertFalse(back1_man.backups_dict)
        file_list = get_file_list(
            self.test_root_back1, name_suffix="events", exclude_dirs=["derivatives"], extensions=[".tsv"]
        )
        self.assertEqual(len(file_list), 3)
        back1_man.create_backup(file_list, backup_name="my_back")
        self.assertTrue(back1_man.backups_dict)
        backup = back1_man.backups_dict["my_back"]
        self.assertEqual(len(backup), len(file_list))
        if os.path.exists(alt_path):
            shutil.rmtree(alt_path)

    def test_bad_data_root(self):
        with self.assertRaises(HedFileError) as context:
            BackupManager("/baloney/Junk")
        self.assertEqual(context.exception.args[0], "NonExistentData")

    def test_constructor_missing_backup(self):
        remove_list = ["back2_miss_json", "back3_miss_back", "back4_miss_file"]
        remove_dirs = [os.path.join(self.test_root_bad_backups, file) for file in remove_list]
        for remove_dir in remove_dirs:
            shutil.rmtree(remove_dir)
        with self.assertRaises(HedFileError) as context:
            BackupManager(self.test_root_bad)
        self.assertEqual(context.exception.code, "MissingBackupFile")

    def test_constructor_missing_json(self):
        remove_list = ["back1_extra", "back3_miss_back", "back4_miss_file"]
        remove_dirs = [os.path.realpath(os.path.join(self.test_root_bad_backups, file)) for file in remove_list]
        for remove_dir in remove_dirs:
            shutil.rmtree(remove_dir)
        with self.assertRaises(HedFileError) as context:
            BackupManager(self.test_root_bad)
        self.assertEqual(context.exception.code, "BadBackupFormat")

    def test_constructor_extra_backup_file(self):
        remove_list = ["back1_extra", "back2_miss_json", "back4_miss_file"]
        remove_dirs = [os.path.realpath(os.path.join(self.test_root_bad_backups, file)) for file in remove_list]
        for remove_dir in remove_dirs:
            shutil.rmtree(remove_dir)
        with self.assertRaises(HedFileError) as context:
            BackupManager(self.test_root_bad)
        self.assertEqual(context.exception.code, "BadBackupFormat")

    def test_create_backup(self):
        test_man = BackupManager(self.test_root)
        file_list = get_file_list(self.test_root)
        self.assertFalse(test_man.get_backup("test_back1"), "create_backup doesn't have the backup before creation")
        return_val1 = test_man.create_backup(file_list, backup_name="test_back1", verbose=False)
        self.assertTrue(return_val1, "create_backup returns true when it has created a backup.")
        backup1 = test_man.get_backup("test_back1")
        self.assertIsInstance(backup1, dict, "create_backup creates a dictionary")
        return_val2 = test_man.create_backup(file_list, backup_name="test_back1", verbose=False)
        self.assertFalse(return_val2, "create_backup returns true when it has created a backup.")

    def test_create_backup_no_name(self):
        test_man = BackupManager(self.test_root)
        self.assertFalse(test_man.backups_dict)
        file_list = get_file_list(self.test_root)
        with patch("sys.stdout", new=io.StringIO()) as fp1:
            return_val1 = test_man.create_backup(file_list, verbose=False)
            self.assertFalse(fp1.getvalue())
        self.assertTrue(return_val1, "create_backup returns true when it has created a backup.")
        backup1 = test_man.get_backup("test_back1")
        self.assertFalse(backup1)
        back = test_man.get_backup(BackupManager.DEFAULT_BACKUP_NAME)
        self.assertIsInstance(back, dict, "create_backup creates a dictionary")
        self.assertTrue(back)
        return_val2 = test_man.create_backup(file_list, backup_name="test_back1", verbose=False)
        self.assertTrue(return_val2, "create_backup returns true when it has created a backup.")

    def test_get_task(self):
        task1 = BackupManager.get_task(["abc", "def"], "temp/myabc.txt")
        self.assertFalse(task1)
        task2 = BackupManager.get_task([], "temp/myabc.txt")
        self.assertFalse(task2)
        task3 = BackupManager.get_task(["abc", "def"], "temp/alpha_key_task_abc.txt")
        self.assertEqual(task3, "abc")


if __name__ == "__main__":
    unittest.main()
